"""Generalized energy harvester data models."""

from collections.abc import Mapping
from enum import Enum
from typing import Annotated
from typing import Any

from pydantic import Field
from pydantic import model_validator
from typing_extensions import Self

from shepherd_core.config import config
from shepherd_core.data_models.base.calibration import CalibrationHarvester
from shepherd_core.data_models.base.content import ContentModel
from shepherd_core.data_models.base.shepherd import ShpModel
from shepherd_core.logger import log
from shepherd_core.testbed_client import tb_client

from .energy_environment import EnergyDType


class AlgorithmDType(str, Enum):
    """Options for choosing a harvesting algorithm."""

    direct = disable = neutral = "neutral"
    """
    Reads an energy environment as is without selecting a harvesting
    voltage.

    Used to play "constant-power" energy environments or simple
    "on-off-patterns". Generally, not useful for virtual source
    emulation.

    Not applicable to real harvesting, only emulation with IVTrace / samples.
    """

    isc_voc = "isc_voc"
    """
    Short Circuit Current, Open Circuit Voltage.

    This is not relevant for emulation, but used to configure recording of
    energy environments.

    This mode samples the two extremes of an IV curve, which may be
    interesting to characterize a transducer/energy environment.

    Not applicable to emulation - only recordable during harvest-recording ATM.
    """

    ivcurve = ivcurves = ivsurface = "ivcurve"
    """
    Used during harvesting to record the full IV surface.

    When configuring the energy environment recording, this algorithm
    records the IV surface by repeatedly recording voltage and current
    while ramping the voltage.

    Cannot be used as output of emulation.
    """

    constant = cv = "cv"
    """
    Harvest energy at a fixed predefined voltage ('voltage_mV').

    For harvesting, this records the IV samples at the specified voltage.
    For emulation, this virtually harvests the IV surface at the specified voltage.

    In addition to constant voltage harvesting, this can be used together
    with the 'feedback_to_hrv' flag to implement a "Capacitor and Diode"
    topology, where the harvesting voltage depends dynamically on the
    capacitor voltage.
    """

    # ci .. constant current -> is this desired?

    mppt_voc = "mppt_voc"
    """
    Emulate a harvester with maximum power point (MPP) tracking based on
    open circuit voltage measurements.

    This MPPT heuristic estimates the MPP as a constant ratio of the open
    circuit voltage.

    Used in conjunction with 'setpoint_n', 'interval_ms', and 'duration_ms'.
    """

    mppt_po = perturb_observe = "mppt_po"
    """
    Emulate a harvester with perturb and observe maximum power point
    tracking.

    This MPPT heuristic adjusts the harvesting voltage by small amounts and
    checks if the power increases. Eventually, the tracking changes the
    direction of adjustments and oscillates around the MPP.
    """

    mppt_opt = optimal = "mppt_opt"
    """
    A theoretical harvester that identifies the MPP by reading it from the
    IV curve during emulation.

    Note that this is not possible for real-world harvesting as the system would
    not know the entire IV curve. In that case a very fast and detailed mppt_po is
    used.
    """


class VirtualHarvesterConfig(ContentModel, title="Config for the Harvester"):
    """The virtual harvester configuration characterizes usage of an energy environment.

    It is used to both harvesting during emulation and to record
    energy environments (sometimes referred to as "harvesting traces").

    For emulation:

    The virtual harvester configuration describes how energy from a recorded
    energy environment is harvested. Typically, the energy environment provides
    an IV-surface, which is a continuous function in three dimensions: voltage,
    current, and time. Based on this surface, the emulation can derive the
    available IV-curve at each point in time. The harvester looks up the current
    that is available (according to the energy environment) from a given
    harvesting voltage. The harvesting voltage may be dynamically chosen by the
    harvester based on the implemented harvesting algorithm, which models
    different real-world harvesters. For example, a maximum power point tracking
    harvester may choose a harvesting voltage as a ratio of the open circuit
    voltage available from the energy environment (or transducer in practice).

    The energy environments are encoded not as a three-dimensional function, but
    as IV tuples over time (sampled at a constant frequency). This originates
    from the technical implementation when recording the IV-surface, where the
    recorder provides the IV-curve by measuring the current for a given voltage
    and ramping the voltage from minimal to maximum.

    For harvest-recordings:

    An energy environment is fully described by the IV surface, which are IV
    curves over time. Shepherd approximates this by sampling the current at
    equidistant steps of a voltage ramp. The VirtualHarvesterConfig is also used
    to parameterize the recording process, typically, it should be configured to
    record a full IV surface, as this contains the full information of the energy
    environment. The postponed harvesting is then performed during emulation.

    However, it is also possible to record a "pre-harvested" energy environment
    by performing the harvesting during recording. This results in a recording
    containing IV samples over time that represent the harvesting voltage
    (chosen by the virtual harvester during recording) and the current available
    from the energy environment for that voltage. Together, these represent the
    power available for harvesting at the time, and during emulation, this power
    can be converted by the input stage (boost converter) to charge the energy
    storage.
    """

    # General Metadata & Ownership -> ContentModel

    algorithm: AlgorithmDType
    """The algorithm determines how the harvester chooses the harvesting voltage.
    """

    samples_n: Annotated[int, Field(ge=8, le=2_000)] = 8
    """How many IV samples are measured for one IV curve.

    The curve is recorded by measuring the el. current available from the
    transducer at equidistant voltage intervals. These voltages are
    probed by ramping between `voltage_min_mV` and `voltage_max_mV` at
    `samples_n` points equally distributed over the voltage range. After
    setting the voltage, the recorder waits for a short period - allowing
    the analog frontend and transducer to settle - before recording the
    harvesting current. This wait duration is influenced by
    `wait_cycles`.

    Selecting all these parameters is a tradeoff between accuracy of the IV
    curve (density of IV samples) and measurement duration, hence the time
    accuracy (density of points) of the IV-surface.

    Only applicable to recording, not used in emulation.

    Used together with `voltage_min_mV`, `voltage_max_mV`, `rising`, and
    `wait_cycles`.
    """

    voltage_mV: Annotated[float, Field(ge=0, le=5_000)] = 2_500
    """The harvesting voltage for constant voltage harvesting.

    Additionally, for Perturb-and-Observe MPPT, this defines the voltage at
    startup.
    """

    voltage_min_mV: Annotated[float, Field(ge=0, le=5_000)] = 0
    """Minimum voltage recorded for the IV curve.

    See `samples_n` for further details.

    In emulation, this can be used to "block" parts of the recorded IV curve
    and not utilize them in the virtual source. However, this is generally
    discouraged as it can result in discontinuities in the curve and is not
    well tested.
    For emulation, this value ideally corresponds to the value of the
    recorded energy environment.
    """

    voltage_max_mV: Annotated[float, Field(ge=0, le=5_000)] = 5_000
    """Maximum voltage sampled for the curve.

    See `voltage_min_mV` and `samples_n`.
    """

    current_limit_uA: Annotated[float, Field(ge=1, le=50_000)] = 50_000
    """
    For MPPT VOC, the open circuit voltage is identified as the el. current
    crosses below this threshold.

    During recording it allows to keep trajectory in special region
    (constant current tracking).
    """

    voltage_step_mV: Annotated[float, Field(ge=1, le=1000000)] | None = None
    """The difference between two adjacent voltage samples.

    This value is implicitly derived from the other ramp parameters:
    (voltage_max_mV - voltage_min_mV) / (samples_n - 1)
    """

    setpoint_n: Annotated[float, Field(ge=0, le=1.0)] = 0.70
    """
    The "Open Circuit Voltage Maximum Power Point Tracker" estimates the MPP
    by taking a constant fraction defined by this parameter of the open
    circuit voltage. For example, if the IV curve shows an open circuit
    voltage of 2V and the setpoint is 0.75, then the harvester selects
    1.5 volts as the harvesting voltage.

    This value is only relevant when 'algorithm == mppt_voc'.
    """

    interval_ms: Annotated[float, Field(ge=0.01, le=1_000_000)] = 100
    """The MPP is repeatedly estimated at fixed intervals defined by this duration.

    Note that the energy environment can still change in between MPP
    estimations, but the harvesting voltage is not updated in between.

    This value is relevant for all MPP algorithms.
    For Perturb and Observe, this value is the wait interval between steps.

    When an energy environment is recorded with `mppt_opt`, the optimal
    harvester is approximated with a very fast Perturb-Observe algorithm,
    where this interval should be set to a very small value.
    When emulating with `mppt_opt`, this value is not relevant as the
    emulation simply picks the maximum power point from the IV-curve.
    """

    duration_ms: Annotated[float, Field(ge=0.01, le=1_000_000)] = 0.1
    """The duration of MPP sampling.

    While performing an MPP sampling every 'interval_ms', the input is
    disconnected to accurately measure the open circuit voltage.

    This value is only relevant for `mppt_voc`.
    """

    rising: bool = True
    """Ramp direction for sampling the IV curve.

    When set to true, sampling starts at the minimum voltage and ramps up to
    the maximum.

    See `samples_n` for further details.
    Not relevant for emulation.
    """

    enable_linear_extrapolation: bool = True
    """
    Because the IV curve is not stored fully in PRU memory but streamed
    sequentially to the PRU, looking up any IV value at any time is not
    possible. However, because the precision of the emulation degrades
    until the relevant IV sample passes by, it can extrapolate the available data
    to estimate the required IV sample.

    Enabling extrapolation can yield a better numeric simulation, especially
    if the harvesting voltage changes rapidly or the IV surface is steep in
    relevant regions. For example, when emulating a capacitor diode
    setup and the current falls at high voltages.

    This value is only relevant for emulation.
    """

    # Underlying recorder
    wait_cycles: Annotated[int, Field(ge=0, le=100)] = 1
    """
    The wait duration to let the analog frontend settle before taking a
    measurement.

    When recording the energy environment, the voltage is set by the
    digital-to-analog-converter. This parameter delays the current
    measurement performed by the analog-to-digital converter to allow the
    harvesting transducer to settle at the defined voltage.

    When recording with `IscVoc`, wait cycles should be added as the analog
    changes are more significant.

    Not relevant for emulation.
    """

    # ⤷ first cycle: ADC-Sampling & DAC-Writing, further steps: waiting

    @model_validator(mode="before")
    @classmethod
    def query_database(cls, values: dict[str, Any]) -> dict[str, Any]:
        values, chain = tb_client.try_completing_model(cls.__name__, values)
        values = tb_client.fill_in_user_data(values)
        if values["name"] == "neutral":
            # TODO: same test is later done in calc_algorithm_num() again
            raise ValueError("Resulting Harvester can't be neutral")
        log.debug("VHrv-Inheritances: %s", chain)

        # post corrections -> should be in separate validator
        cal = CalibrationHarvester()  # TODO: as argument?
        c_limit = values.get("current_limit_uA", 50_000)  # cls.current_limit_uA)
        values["current_limit_uA"] = max(10**6 * cal.adc_C_Hrv.raw_to_si(4), c_limit)

        if values.get("voltage_step_mV") is None:
            # algo includes min & max!
            v_max = values.get("voltage_max_mV", 5_000)  # cls.voltage_max_mV)
            v_min = values.get("voltage_min_mV", 0)  # cls.voltage_min_mV)
            samples_n = values.get("samples_n", 8)  # cls.samples_n) TODO
            values["voltage_step_mV"] = abs(v_max - v_min) / (samples_n - 1)

        values["voltage_step_mV"] = max(
            10**3 * cal.dac_V_Hrv.raw_to_si(4), values["voltage_step_mV"]
        )

        return values

    @model_validator(mode="after")
    def post_validation(self) -> Self:
        if self.voltage_min_mV > self.voltage_max_mV:
            raise ValueError("Voltage minimum > max")
        if self.voltage_mV < self.voltage_min_mV:
            raise ValueError("Voltage below minimum")
        if self.voltage_mV > self.voltage_max_mV:
            raise ValueError("Voltage above maximum")

        return self

    def calc_hrv_mode(self, *, for_emu: bool) -> int:
        return 1 * int(for_emu) + 2 * self.rising + 4 * self.enable_linear_extrapolation

    def calc_algorithm_num(self, *, for_emu: bool) -> int:
        num: int = ALGO_TO_NUM.get(self.algorithm, ALGO_TO_NUM["neutral"])
        if for_emu and self.get_datatype() != EnergyDType.ivsample:
            msg = (
                f"[{self.name}] Select valid harvest-algorithm for emulator, "
                f"current usage = {self.algorithm}"
            )
            raise ValueError(msg)
        if not for_emu and num < ALGO_TO_NUM["isc_voc"]:
            msg = (
                f"[{self.name}] Select valid harvest-algorithm for harvester, "
                f"current usage = {self.algorithm}"
            )
            raise ValueError(msg)
        return num

    def calc_timings_ms(self, *, for_emu: bool) -> tuple[float, float]:
        """factor-in model-internal timing-constraints."""
        window_length = self.samples_n * (1 + self.wait_cycles)
        time_min_ms = (1 + self.wait_cycles) * 1_000 / config.SAMPLERATE_SPS
        if for_emu:
            window_ms = window_length * 1_000 / config.SAMPLERATE_SPS
            time_min_ms = max(time_min_ms, window_ms)

        interval_ms = min(max(self.interval_ms, time_min_ms), 1_000_000)
        duration_ms = min(max(self.duration_ms, time_min_ms), interval_ms)
        _ratio = (duration_ms / interval_ms) / (self.duration_ms / self.interval_ms)
        if (_ratio - 1) > 0.1:
            log.debug(
                "Ratio between interval & duration has changed "
                "more than 10%% due to constraints (%.4f)",
                _ratio,
            )
        return interval_ms, duration_ms

    def get_datatype(self) -> EnergyDType:
        return ALGO_TO_DTYPE[self.algorithm]

    def calc_window_size(
        self,
        dtype_in: EnergyDType | None = None,
        *,
        for_emu: bool,
    ) -> int:
        if not for_emu:
            # TODO: should be named 'for_ivcurve_recording'
            # TODO: add extra variable to distinguish step_count
            #       and window_size (currently mixed together)
            # only used by ivcurve algo (in ADC-Mode)
            return self.samples_n

        if dtype_in is None:
            dtype_in = self.get_datatype()

        if dtype_in == EnergyDType.ivcurve:
            return self.samples_n * (1 + self.wait_cycles)
        if dtype_in == EnergyDType.ivsample:
            return 0
        if dtype_in == EnergyDType.isc_voc:
            return 2 * (1 + self.wait_cycles)
        raise NotImplementedError


u32 = Annotated[int, Field(ge=0, lt=2**32)]


# Currently implemented harvesters
# NOTE: numbers have meaning and will be tested ->
# - harvesting on "neutral" is not possible - direct pass-through
# - emulation with "ivcurve" or lower is also resulting in Error
# - "_opt" has its own algo for emulation, but is only a fast mppt_po for harvesting
ALGO_TO_NUM: Mapping[str, int] = {
    "neutral": 2**0,
    "isc_voc": 2**3,
    "ivcurve": 2**4,
    "cv": 2**8,
    # is "ci" desired?
    "mppt_voc": 2**12,
    "mppt_po": 2**13,
    "mppt_opt": 2**14,
}

ALGO_TO_DTYPE: Mapping[str, EnergyDType] = {
    "neutral": EnergyDType.ivsample,
    "isc_voc": EnergyDType.isc_voc,
    "ivcurve": EnergyDType.ivcurve,
    "cv": EnergyDType.ivsample,
    "mppt_voc": EnergyDType.ivsample,
    "mppt_po": EnergyDType.ivsample,
    "mppt_opt": EnergyDType.ivsample,
}


class HarvesterPRUConfig(ShpModel):
    """Map settings-list to internal state-vars struct HarvesterConfig for PRU.

    NOTE:
      - yaml is based on si-units like nA, mV, ms, uF
      - c-code and py-copy is using nA, uV, ns, nF, fW, raw
      - ordering is intentional and in sync with shepherd/commons.h.
    """

    algorithm: u32
    hrv_mode: u32
    window_size: u32
    voltage_uV: u32
    voltage_min_uV: u32
    voltage_max_uV: u32
    voltage_step_uV: u32
    """ ⤷ for window-based algo like ivcurve"""
    current_limit_nA: u32
    """ ⤷ lower bound to detect zero current"""
    setpoint_n8: u32
    interval_n: u32
    """ ⤷ between measurements"""
    duration_n: u32
    """ ⤷ of measurement"""
    wait_cycles_n: u32
    """ ⤷ for DAC to settle"""

    @classmethod
    def from_vhrv(
        cls,
        data: VirtualHarvesterConfig,
        dtype_in: EnergyDType | None = EnergyDType.ivsample,
        window_size: u32 | None = None,
        voltage_step_V: float | None = None,
        *,
        for_emu: bool = False,
    ) -> Self:
        if isinstance(dtype_in, str):
            dtype_in = EnergyDType[dtype_in]
        if for_emu and dtype_in not in {EnergyDType.ivsample, EnergyDType.ivcurve}:
            raise NotImplementedError

        if for_emu and dtype_in == EnergyDType.ivcurve and voltage_step_V is None:
            raise ValueError(
                "For correct emulation specify voltage_step used by harvester "
                "e.g. via file_src.get_voltage_step()"
            )

        if for_emu and dtype_in == EnergyDType.ivcurve and window_size is None:
            raise ValueError(
                "For correct emulation specify window_size used by harvester "
                "e.g. via file_src.get_window_size()"
            )

        interval_ms, duration_ms = data.calc_timings_ms(for_emu=for_emu)
        window_size = (
            window_size
            if window_size is not None
            else data.calc_window_size(dtype_in, for_emu=for_emu)
        )
        if voltage_step_V is not None:
            voltage_step_mV = 1e3 * voltage_step_V
        elif data.voltage_step_mV is not None:
            voltage_step_mV = data.voltage_step_mV
        else:
            raise ValueError(
                "For correct emulation specify voltage_step used by harvester "
                "e.g. via file_src.get_voltage_step()"
            )

        return cls(
            algorithm=data.calc_algorithm_num(for_emu=for_emu),
            hrv_mode=data.calc_hrv_mode(for_emu=for_emu),
            window_size=window_size,
            voltage_uV=round(data.voltage_mV * 10**3),
            voltage_min_uV=round(data.voltage_min_mV * 10**3),
            voltage_max_uV=round(data.voltage_max_mV * 10**3),
            voltage_step_uV=round(voltage_step_mV * 10**3),
            current_limit_nA=round(data.current_limit_uA * 10**3),
            setpoint_n8=round(min(255, data.setpoint_n * 2**8)),
            interval_n=round(interval_ms * config.SAMPLERATE_SPS * 1e-3),
            duration_n=round(duration_ms * config.SAMPLERATE_SPS * 1e-3),
            wait_cycles_n=data.wait_cycles,
        )



"""Simulation model of the virtual source."""

from .target_model import ConstantCurrentTarget
from .target_model import ConstantPowerTarget
from .target_model import ResistiveTarget
from .virtual_converter_model import PruCalibration
from .virtual_converter_model import VirtualConverterModel
from .virtual_harvester_model import VirtualHarvesterModel
from .virtual_harvester_simulation import simulate_harvester
from .virtual_source_model import VirtualSourceModel
from .virtual_source_simulation import simulate_source
from .virtual_storage_model import VirtualStorageModel
from .virtual_storage_simulator import StorageSimulator

__all__ = [
    "ConstantCurrentTarget",
    "ConstantPowerTarget",
    "PruCalibration",
    "ResistiveTarget",
    "StorageSimulator",
    "VirtualConverterModel",
    "VirtualHarvesterModel",
    "VirtualSourceModel",
    "VirtualStorageModel",
    "simulate_harvester",
    "simulate_source",
]



"""Generalized virtual source data models."""

from typing import Annotated
from typing import Any

from pydantic import Field
from pydantic import model_validator
from typing_extensions import Self

from shepherd_core.config import config
from shepherd_core.data_models.base.content import ContentModel
from shepherd_core.data_models.base.shepherd import ShpModel
from shepherd_core.logger import log
from shepherd_core.testbed_client import tb_client

from .energy_environment import EnergyDType
from .virtual_harvester_config import HarvesterPRUConfig
from .virtual_harvester_config import VirtualHarvesterConfig
from .virtual_storage_config import VirtualStorageConfig

# Custom Types
LUT_SIZE: int = 12
NormedNum = Annotated[float, Field(ge=0.0, le=1.0)]
LUT1D = Annotated[list[NormedNum], Field(min_length=LUT_SIZE, max_length=LUT_SIZE)]
LUT2D = Annotated[list[LUT1D], Field(min_length=LUT_SIZE, max_length=LUT_SIZE)]

# defaults (pre-init complex types for improved perf) TODO: is documentation still fine?
vhrv_mppt_opt = VirtualHarvesterConfig(name="mppt_opt")


class VirtualSourceConfig(ContentModel, title="Config for the virtual Source"):
    """The vSrc uses the energy environment (file) for supplying the Target Node.

    If not already done, the energy will be harvested and
    then converted during the experiment.

    The converter-stage is software defined and offers:
    - buck-boost-combinations,
    - a simple diode + resistor and
    - an intermediate storage capacitor.
    """

    # TODO: I,V,R should be in regular unit (V, A, Ohm)

    # General Metadata & Ownership -> ContentModel

    enable_boost: bool = False
    """ ⤷ if false -> v_intermediate = v_input, output-switch-hysteresis is still usable"""
    enable_buck: bool = False
    """ ⤷ if false -> v_output = v_intermediate"""
    enable_feedback_to_hrv: bool = False
    """ Source can control a cv-harvester for ivsurface.
     Feedback is essential for some harvesters, i.e. diode-circuitry.
    """

    interval_startup_delay_drain_ms: Annotated[float, Field(ge=0, le=10_000)] = 0
    """ ⤷ Model begins running, but Target is not draining the storage capacitor
    until this delay is over.
    """

    harvester: VirtualHarvesterConfig = vhrv_mppt_opt
    """ ⤷ Only active / needed if input is ivsurface. """

    V_input_max_mV: Annotated[float, Field(ge=0, le=10_000)] = 10_000
    """ ⤷ Maximum input Voltage [mV] -> will be clipped."""
    I_input_max_mA: Annotated[float, Field(ge=0, le=4.29e3)] = 4_200
    """ ⤷ Maximum input Current [mA] -> will be clipped."""
    V_input_drop_mV: Annotated[float, Field(ge=0, le=4.29e6)] = 0
    """ ⤷ simulate voltage drop for input-diode or LDO."""
    R_input_mOhm: Annotated[float, Field(ge=0, le=4.29e6)] = 0
    """ ⤷ resistance only active with disabled boost, range [1 mOhm; 1MOhm]"""

    storage: VirtualStorageConfig | None = None
    """ ⤷ primary intermediate energy storage between boost- and buck-converter stage.
    Selecting "None" disables the storage and directly connects input to output.
    """
    V_intermediate_enable_output_threshold_mV: Annotated[float, Field(ge=0, le=10_000)] = 1
    """ ⤷ target gets connected (hysteresis-combo with next value)"""
    V_intermediate_disable_output_threshold_mV: Annotated[float, Field(ge=0, le=10_000)] = 0
    """ ⤷ target gets disconnected"""
    interval_check_thresholds_ms: Annotated[float, Field(ge=0, le=4.29e3)] = 0
    """ ⤷ some ICs (BQ) check every 64 ms if output should be disconnected"""
    # TODO: add intervals for input-disable, output-disable & power-good-signal

    V_pwr_good_enable_threshold_mV: Annotated[float, Field(ge=0, le=10_000)] = 2_800
    """ pwr-good: target is informed on output-pin (hysteresis)
     -> reference is the intermediate voltage """
    V_pwr_good_disable_threshold_mV: Annotated[float, Field(ge=0, le=10_000)] = 2200
    immediate_pwr_good_signal: bool = True
    """ ⤷ 1: activate instant schmitt-trigger, 0: stay in interval for checking thresholds"""

    C_output_uF: Annotated[float, Field(ge=0, le=4.29e6)] = 1.0
    """
    final (always last) stage to compensate undetectable current spikes when
    enabling power for target
    """
    # TODO: C_output is handled internally as delta-V, but should be a I_transient
    #       that makes it visible in simulation as additional i_out_drain
    # TODO: potential weakness, ACD lowpass is capturing transient,
    #       but energy is LOST with this model

    # Extra
    V_output_log_gpio_threshold_mV: Annotated[float, Field(ge=0, le=4.29e6)] = 1_400
    """ ⤷ minimum voltage threshold needed to enable recording changes in gpio-bank"""

    # Boost Converter
    V_input_boost_threshold_mV: Annotated[float, Field(ge=0, le=10_000)] = 0
    """ ⤷ minimum input-voltage for the boost converter to work"""
    V_intermediate_max_mV: Annotated[float, Field(ge=0, le=10_000)] = 10_000
    """ ⤷ threshold for shutting off boost converter """

    LUT_input_efficiency: LUT2D = 12 * [12 * [1.00]]
    """ ⤷ rows are current -> first row a[V=0][:]

    input-LUT[12][12] depending on array[inp_voltage][log(inp_current)],
    influence of cap-voltage is not implemented
    """

    LUT_input_V_min_log2_uV: Annotated[int, Field(ge=0, le=20)] = 0
    """ ⤷ i.e. 2^7 = 128 uV -> LUT[0][:] is for inputs < 128 uV"""
    LUT_input_I_min_log2_nA: Annotated[int, Field(ge=1, le=20)] = 1
    """ ⤷ i.e. 2^8 = 256 nA -> LUT[:][0] is for inputs < 256 nA"""

    # Buck Converter
    V_output_mV: Annotated[float, Field(ge=0, le=5_000)] = 2_400
    """ Fixed Voltage of Buck-Converter.
    (as long as Input is > Output + Drop-Voltage)
    """
    V_buck_drop_mV: Annotated[float, Field(ge=0, le=5_000)] = 0
    """ ⤷ simulate LDO / diode minimum voltage differential or output-diode"""

    LUT_output_efficiency: LUT1D = 12 * [1.00]
    """ ⤷ array[12] depending on output_current, In- & Output is linear."""
    LUT_output_I_min_log2_nA: Annotated[int, Field(ge=1, le=20)] = 1
    """ ⤷ i.e. 2^8 = 256 nA -> LUT[0] is for inputs < 256 nA,
    see notes on LUT_input for explanation
    """

    @model_validator(mode="before")
    @classmethod
    def query_database(cls, values: dict[str, Any]) -> dict[str, Any]:
        values, chain = tb_client.try_completing_model(cls.__name__, values)
        values = tb_client.fill_in_user_data(values)
        log.debug("VSrc-Inheritances: %s", chain)
        # TODO: most "internal states" should be corrected here

        return values

    @model_validator(mode="after")
    def post_validation(self) -> Self:
        # trigger stricter test of harv-parameters
        HarvesterPRUConfig.from_vhrv(self.harvester, for_emu=True)
        # TODO: enable threshold < mid_max
        # TODO: mid_max < mid_soc1
        return self

    def calc_internal_states(self) -> dict:
        """Update the model-states for the capacitor and other elements.

        This also compensates for current-surge of real capacitors
        when the converter gets turned on:

        - surges are hard to detect & record
        - this can be const value, because
        - the converter always turns on with "V_storage_enable_threshold_uV".

        TODO: currently neglecting delay after disabling converter, boost
        only has simpler formula, second enabling when V_Cap >= V_out

        Math behind this calculation:

        - Energy-Change Storage Cap ->  E_new = E_old - E_output
        - with Energy of a Cap 	    -> 	E_x = C_x * V_x^2 / 2
        - combine formulas 		    ->  C_store * V_store_new^2 / 2 =
          C_store * V_store_old^2 / 2 - C_out * V_out^2 / 2
        - convert formula to V_new 	->	V_store_new^2 =
          V_store_old^2 - (C_out / C_store) * V_out^2
        - convert into dV	 	        ->	dV = V_store_new - V_store_old
        - in case of V_cap = V_out 	-> 	dV = V_store_old * (sqrt(1 - C_out / C_store) - 1)

        Note: dV values will be reversed (negated), because dV is always negative (Voltage drop)
        """
        values = {}
        if (self.storage is not None) and self.C_output_uF > 0:
            # first case: storage cap outside of en/dis-thresholds
            v_old = self.V_intermediate_enable_output_threshold_mV
            v_out = self.V_output_mV
            c_store = self.storage.capacity_in_uF
            c_out = self.C_output_uF
            dV_output_en_thrs_mV = v_old - pow(
                pow(v_old, 2) - (c_out / c_store) * pow(v_out, 2),
                0.5,
            )

            # second case: storage cap below v_out (only different for enabled buck),
            #              enable when >= v_out
            # v_enable is either bucks v_out or same dV-Value is calculated a second time
            dV_output_imed_low_mV = v_out * (1 - pow(1 - c_out / c_store, 0.5))
        else:
            dV_output_en_thrs_mV = 0
            dV_output_imed_low_mV = 0

        if self.enable_boost and self.storage is not None:
            # TODO: storage could have maximum at a different SoC, is this needed at all?
            values["V_mid_max_mV"] = min(
                self.V_intermediate_max_mV,
                1e3 * self.storage.calc_V_OC(SoC=1.0),
                10_000,
            )
        else:
            values["V_mid_max_mV"] = self.V_intermediate_max_mV

        # protect from complex solutions (non valid input combinations)
        if not (isinstance(dV_output_en_thrs_mV, (int, float)) and (dV_output_en_thrs_mV >= 0)):
            dV_output_en_thrs_mV = 0
        if not (isinstance(dV_output_imed_low_mV, (int, float)) and (dV_output_imed_low_mV >= 0)):
            log.warning("VSrc: C_output shouldn't be larger than C_intermediate")
            dV_output_imed_low_mV = 0

        # decide which hysteresis-thresholds to use for buck-converter
        if self.enable_buck > 0:
            V_pre_output_mV = self.V_output_mV + self.V_buck_drop_mV

            if self.V_intermediate_enable_output_threshold_mV > V_pre_output_mV:
                values["dV_mid_enable_output_mV"] = dV_output_en_thrs_mV
                values["V_mid_enable_output_threshold_mV"] = (
                    self.V_intermediate_enable_output_threshold_mV
                )

            else:
                values["dV_mid_enable_output_mV"] = dV_output_imed_low_mV
                values["V_mid_enable_output_threshold_mV"] = (
                    V_pre_output_mV + values["dV_mid_enable_output_mV"]
                )

            if self.V_intermediate_disable_output_threshold_mV > V_pre_output_mV:
                values["V_mid_disable_output_threshold_mV"] = (
                    self.V_intermediate_disable_output_threshold_mV
                )
            else:
                values["V_mid_disable_output_threshold_mV"] = V_pre_output_mV

        else:
            values["dV_mid_enable_output_mV"] = dV_output_en_thrs_mV
            values["V_mid_enable_output_threshold_mV"] = (
                self.V_intermediate_enable_output_threshold_mV
            )
            values["V_mid_disable_output_threshold_mV"] = (
                self.V_intermediate_disable_output_threshold_mV
            )
        return values

    def calc_converter_mode(self, dtype_in: EnergyDType, *, log_intermediate_node: bool) -> int:
        """Assembles bitmask from discrete values.

        log_intermediate_node: record / log virtual intermediate (cap-)voltage and
        -current (out) instead of output-voltage and -current
        """
        enable_storage = self.storage is not None
        enable_boost = self.enable_boost and enable_storage
        if enable_boost != self.enable_boost:
            log.warning("VSrc - boost was disabled due to missing storage capacitor!")
        enable_feedback = (
            self.enable_feedback_to_hrv
            and enable_storage
            and not enable_boost
            and dtype_in == EnergyDType.ivcurve
        )
        if enable_feedback != self.enable_feedback_to_hrv:
            reason = "enabled boost, " if enable_boost else ""
            reason += "" if dtype_in == EnergyDType.ivcurve else "input not ivcurve, "
            reason += "" if enable_storage else "no storage capacitor"
            log.warning("VSRC - feedback to harvester was disabled! Reasons: %s", reason)
        return (
            1 * int(enable_storage)
            + 2 * int(enable_boost)
            + 4 * int(self.enable_buck)
            + 8 * int(log_intermediate_node)
            + 16 * int(enable_feedback)
        )


u32 = Annotated[int, Field(ge=0, lt=2**32)]
u8 = Annotated[int, Field(ge=0, lt=2**8)]
lut_i = Annotated[
    list[Annotated[list[u8], Field(min_length=LUT_SIZE, max_length=LUT_SIZE)]],
    Field(
        min_length=LUT_SIZE,
        max_length=LUT_SIZE,
    ),
]
lut_o = Annotated[list[u32], Field(min_length=LUT_SIZE, max_length=LUT_SIZE)]


class ConverterPRUConfig(ShpModel):
    """Map settings-list to internal state-vars struct ConverterConfig.

    NOTE:
      - yaml is based on si-units like nA, mV, ms, uF
      - c-code and py-copy is using nA, uV, ns, nF, fW, raw
      - ordering is intentional and in sync with shepherd/commons.h
    """

    converter_mode: u32
    interval_startup_delay_drain_n: u32

    V_input_max_uV: u32
    I_input_max_nA: u32
    V_input_drop_uV: u32
    R_input_kOhm_n22: u32
    # ⤷ TODO: possible optimization: n32 (range 1uOhm to 1 kOhm) is easier to calc in pru

    V_mid_enable_output_threshold_uV: u32
    V_mid_disable_output_threshold_uV: u32
    dV_mid_enable_output_uV: u32
    interval_check_thresholds_n: u32

    V_pwr_good_enable_threshold_uV: u32
    V_pwr_good_disable_threshold_uV: u32
    immediate_pwr_good_signal: u32

    V_output_log_gpio_threshold_uV: u32

    V_input_boost_threshold_uV: u32
    V_mid_max_uV: u32

    V_output_uV: u32
    V_buck_drop_uV: u32

    LUT_input_V_min_log2_uV: u32
    LUT_input_I_min_log2_nA: u32
    LUT_output_I_min_log2_nA: u32
    LUT_inp_efficiency_n8: lut_i
    LUT_out_inv_efficiency_n4: lut_o

    @classmethod
    def from_vsrc(
        cls,
        data: VirtualSourceConfig,
        dtype_in: EnergyDType = EnergyDType.ivsample,
        *,
        log_intermediate_node: bool = False,
    ) -> Self:
        states = data.calc_internal_states()
        return cls(
            # General
            converter_mode=data.calc_converter_mode(
                dtype_in, log_intermediate_node=log_intermediate_node
            ),
            interval_startup_delay_drain_n=round(
                data.interval_startup_delay_drain_ms * config.SAMPLERATE_SPS * 1e-3
            ),
            V_input_max_uV=round(data.V_input_max_mV * 1e3),
            I_input_max_nA=round(data.I_input_max_mA * 1e6),
            V_input_drop_uV=round(data.V_input_drop_mV * 1e3),
            R_input_kOhm_n22=round(data.R_input_mOhm * (1e-6 * 2**22)),
            V_mid_enable_output_threshold_uV=round(
                states["V_mid_enable_output_threshold_mV"] * 1e3
            ),
            V_mid_disable_output_threshold_uV=round(
                states["V_mid_disable_output_threshold_mV"] * 1e3
            ),
            dV_mid_enable_output_uV=round(states["dV_mid_enable_output_mV"] * 1e3),
            interval_check_thresholds_n=round(
                data.interval_check_thresholds_ms * config.SAMPLERATE_SPS * 1e-3
            ),
            V_pwr_good_enable_threshold_uV=round(data.V_pwr_good_enable_threshold_mV * 1e3),
            V_pwr_good_disable_threshold_uV=round(data.V_pwr_good_disable_threshold_mV * 1e3),
            immediate_pwr_good_signal=data.immediate_pwr_good_signal,
            V_output_log_gpio_threshold_uV=round(data.V_output_log_gpio_threshold_mV * 1e3),
            # Boost-Converter
            V_input_boost_threshold_uV=round(data.V_input_boost_threshold_mV * 1e3),
            V_mid_max_uV=round(states["V_mid_max_mV"] * 1e3),
            # Buck-Converter
            V_output_uV=round(data.V_output_mV * 1e3),
            V_buck_drop_uV=round(data.V_buck_drop_mV * 1e3),
            # LUTs
            LUT_input_V_min_log2_uV=data.LUT_input_V_min_log2_uV,
            LUT_input_I_min_log2_nA=data.LUT_input_I_min_log2_nA - 1,  # sub-1 due to later log2-op
            LUT_output_I_min_log2_nA=data.LUT_output_I_min_log2_nA - 1,  # sub-1 due to later log2
            LUT_inp_efficiency_n8=[
                [min(255, round(256 * ival)) for ival in il] for il in data.LUT_input_efficiency
            ],
            LUT_out_inv_efficiency_n4=[
                min((2**14), round((2**4) / value)) if (value > 0) else (2**14)
                for value in data.LUT_output_efficiency
            ],
        )



"""Demonstrates decoding of a sampled waveform to UART-symbols."""

from pathlib import Path
from timeit import timeit

from shepherd_core.decoder_waveform import Uart
from shepherd_core.logger import log

# file captured with logic analyzer, 15.5k events (2700 symbols, 61 lines)
trace = Path(__file__).parent / "uart_raw2.csv"
uwd = Uart(trace)

sym = uwd.get_symbols()
lne = uwd.get_lines()
txt = uwd.get_text()
log.info(txt)

do_analysis = False
if do_analysis:
    l0 = timeit("Uart(trace)", globals=globals(), number=1000)
    l1 = timeit("uwd.get_symbols(force_redo=True)", globals=globals(), number=100)
    l2 = timeit("uwd.get_lines(force_redo=True)", globals=globals(), number=1000)
    l3 = timeit("uwd.get_text(force_redo=True)", globals=globals(), number=1000)
    print("t_init\t", l0)
    print("t_symb\t", l1 * 10)
    print("t_line\t", l2)
    print("t_text\t", l3)
    # Results:
    # t_init  5.8   [ms/run]
    # t_symb  70.4  [!!!!!]
    # t_line  3.9
    # t_text  0.1



"""this is ported py-version of the pru-code.

Goals:

- stay close to original code-base
- offer a comparison for the tests
- step 1 to a virtualization of emulation

NOTE1: DO NOT OPTIMIZE -> stay close to original c-code-base
NOTE2: adc-harvest-routines are not part of this model (virtual_harvester lines 66:289)

Compromises:

- Py has to map the settings-list to internal vars -> is kernel-task
- Python has no static vars -> FName_reset is handling the class-vars

"""

from shepherd_core.data_models.content.virtual_harvester_config import HarvesterPRUConfig
from shepherd_core.logger import log


class VirtualHarvesterModel:
    """Ported python version of the pru vHrv."""

    HRV_IVCURVE: int = 2**4
    HRV_CV: int = 2**8
    HRV_MPPT_VOC: int = 2**12
    HRV_MPPT_PO: int = 2**13
    HRV_MPPT_OPT: int = 2**14

    def __init__(self, cfg: HarvesterPRUConfig) -> None:
        self._cfg: HarvesterPRUConfig = cfg

        # INIT global vars: shared states
        self.voltage_set_uV: int = self._cfg.voltage_uV + 1

        self.is_emu: bool = bool(self._cfg.hrv_mode & (2**0))
        if not self.is_emu:
            log.warning(
                "This VSrc-config is not meant for emulation-mode -> activate 'is_emu' flag."
            )

        if self._cfg.interval_n > 2 * self._cfg.window_size:
            self.interval_step = self._cfg.interval_n - (2 * self._cfg.window_size)
        else:
            self.interval_step = 2**30
        # ⤷ intake two ivcurves before overflow / reset

        self.is_rising: bool = bool(self._cfg.hrv_mode & (2**1))

        # PO-Relevant, iv & adc
        self.volt_step_uV: int = self._cfg.voltage_step_uV

        # adc_ivcurve
        # self.settle_steps: int = 0  # noqa: ERA001

        # adc_mppt_po
        # self.power_last_raw: int = 0 # noqa: ERA001

        # globals for iv_cv
        self.voltage_hold: int = 0
        self.current_hold: int = 0
        self.voltage_step_x4_uV: int = 4 * self._cfg.voltage_step_uV
        self.age_max: int = 2 * self._cfg.window_size

        self.voc_now: int = self._cfg.voltage_max_uV
        self.voc_nxt: int = self._cfg.voltage_max_uV
        self.voc_min: int = max(1000, self._cfg.voltage_min_uV)

        self.lin_extrapolation: bool = bool(self._cfg.hrv_mode & (2**2))

        # INIT static vars: CV
        self.voltage_last: int = 0
        self.current_last: int = 0
        self.voltage_delta: int = 0
        self.current_delta: int = 0
        self.compare_last: int = 0

        # INIT static vars: VOC
        self.age_now: int = 0
        self.age_nxt: int = 0

        # INIT static vars: PO
        self.power_last: int = 0

        # INIT static vars: OPT
        # already done in VOC: age_now, age_nxt
        self.voltage_now: int = 0
        self.current_now: int = 0
        self.voltage_nxt: int = 0
        self.current_nxt: int = 0
        self.power_now: int = 0
        self.power_nxt: int = 0

    def ivcurve_sample(self, _voltage_uV: int, _current_nA: int) -> tuple[int, int]:
        if self._cfg.window_size <= 1:
            return _voltage_uV, _current_nA
        if self._cfg.algorithm >= self.HRV_MPPT_OPT:
            return self.ivcurve_2_mppt_opt(_voltage_uV, _current_nA)
        if self._cfg.algorithm >= self.HRV_MPPT_PO:
            return self.ivcurve_2_mppt_po(_voltage_uV, _current_nA)
        if self._cfg.algorithm >= self.HRV_MPPT_VOC:
            return self.ivcurve_2_mppt_voc(_voltage_uV, _current_nA)
        if self._cfg.algorithm >= self.HRV_CV:
            return self.ivcurve_2_cv(_voltage_uV, _current_nA)
        # next line is only implied in C
        return _voltage_uV, _current_nA

    def ivcurve_2_cv(self, _voltage_uV: int, _current_nA: int) -> tuple[int, int]:
        compare_now = _voltage_uV < self.voltage_set_uV
        step_size_now = abs(_voltage_uV - self.voltage_last)
        distance_now = abs(_voltage_uV - self.voltage_set_uV)
        distance_last = abs(self.voltage_last - self.voltage_set_uV)

        if compare_now != self.compare_last and step_size_now < self.voltage_step_x4_uV:
            if distance_now < distance_last and distance_now < self.voltage_step_x4_uV:
                self.voltage_hold = _voltage_uV
                self.current_hold = _current_nA
                self.voltage_delta = _voltage_uV - self.voltage_last
                self.current_delta = _current_nA - self.current_last
            elif distance_last < distance_now and distance_last < self.voltage_step_x4_uV:
                self.voltage_hold = self.voltage_last
                self.current_hold = self.current_last
                self.voltage_delta = _voltage_uV - self.voltage_last
                self.current_delta = _current_nA - self.current_last
        elif self.lin_extrapolation:
            # apply the proper delta if needed
            # TODO: C-Code differs here slightly, but only to handle unsigned int
            if (self.voltage_hold < self.voltage_set_uV) == (self.voltage_delta > 0):
                if self.voltage_hold > -self.voltage_delta:
                    self.voltage_hold += self.voltage_delta
                else:
                    self.voltage_hold = 0
                if self.current_hold > -self.current_delta:
                    self.current_hold += self.current_delta
                else:
                    self.current_hold = 0
            else:
                if self.voltage_hold > self.voltage_delta:
                    self.voltage_hold -= self.voltage_delta
                else:
                    self.voltage_hold = 0
                if self.current_hold > self.current_delta:
                    self.current_hold -= self.current_delta
                else:
                    self.current_hold = 0

        self.voltage_last = _voltage_uV
        self.current_last = _current_nA
        self.compare_last = compare_now
        return self.voltage_hold, self.current_hold

    def ivcurve_2_mppt_voc(self, _voltage_uV: int, _current_nA: int) -> tuple[int, int]:
        self.interval_step = self.interval_step + 1
        if self.interval_step >= self._cfg.interval_n:
            self.interval_step = 0
        self.age_nxt += 1
        self.age_now += 1

        if (
            (_current_nA < self._cfg.current_limit_nA)
            and (_voltage_uV < self.voc_nxt)
            and (_voltage_uV >= self.voc_min)
            and (_voltage_uV <= self._cfg.voltage_max_uV)
        ):
            self.voc_nxt = _voltage_uV
            self.age_nxt = 0

        if (self.age_now > self.age_max) or (self.voc_nxt <= self.voc_now):
            self.age_now = self.age_nxt
            self.voc_now = self.voc_nxt
            self.age_nxt = 0
            self.voc_nxt = self._cfg.voltage_max_uV

        _voltage_uV, _current_nA = self.ivcurve_2_cv(_voltage_uV, _current_nA)
        if self.interval_step < self._cfg.duration_n:
            self.voltage_set_uV = self._cfg.voltage_max_uV
            _current_nA = 0
        elif self.interval_step == self._cfg.duration_n:
            self.voltage_set_uV = int(self.voc_now * self._cfg.setpoint_n8 / 256)
            _current_nA = 0
        return _voltage_uV, _current_nA

    def ivcurve_2_mppt_po(self, _voltage_uV: int, _current_nA: int) -> tuple[int, int]:
        self.interval_step = self.interval_step + 1
        if self.interval_step >= self._cfg.interval_n:
            self.interval_step = 0

        _voltage_uV, _current_nA = self.ivcurve_2_cv(_voltage_uV, _current_nA)

        if self.interval_step == 0:
            power_now = _voltage_uV * _current_nA
            if power_now > self.power_last:
                if self.is_rising:
                    self.voltage_set_uV += self.volt_step_uV
                else:
                    self.voltage_set_uV -= self.volt_step_uV
                self.volt_step_uV *= 2
            elif (power_now <= 0) and (self.voltage_set_uV > 0):
                self.is_rising = True
                self.volt_step_uV = self._cfg.voltage_step_uV
                self.voltage_set_uV -= self.voltage_step_x4_uV
            else:
                self.is_rising = not self.is_rising
                self.volt_step_uV = self._cfg.voltage_step_uV
                if self.is_rising:
                    self.voltage_set_uV += self.volt_step_uV
                else:
                    self.voltage_set_uV -= self.volt_step_uV

            self.power_last = power_now

            if self.voltage_set_uV >= self._cfg.voltage_max_uV:
                self.voltage_set_uV = self._cfg.voltage_max_uV
                self.is_rising = False
                self.volt_step_uV = self._cfg.voltage_step_uV
            if self.voltage_set_uV <= self._cfg.voltage_min_uV:
                self.voltage_set_uV = self._cfg.voltage_min_uV
                self.is_rising = True
                self.volt_step_uV = self._cfg.voltage_step_uV
            if self.voltage_set_uV < self._cfg.voltage_step_uV:
                self.voltage_set_uV = self._cfg.voltage_step_uV
                self.is_rising = True
                self.volt_step_uV = self._cfg.voltage_step_uV

        return _voltage_uV, _current_nA

    def ivcurve_2_mppt_opt(self, _voltage_uV: int, _current_nA: int) -> tuple[int, int]:
        self.age_now += 1
        self.age_nxt += 1

        power_fW = _voltage_uV * _current_nA
        if (
            (power_fW >= self.power_nxt)
            and (_voltage_uV >= self._cfg.voltage_min_uV)
            and (_voltage_uV <= self._cfg.voltage_max_uV)
        ):
            self.age_nxt = 0
            self.power_nxt = power_fW
            self.voltage_nxt = _voltage_uV
            self.current_nxt = _current_nA

        if (self.age_now > self.age_max) or (self.power_nxt >= self.power_now):
            self.age_now = self.age_nxt
            self.power_now = self.power_nxt
            self.voltage_now = self.voltage_nxt
            self.current_now = self.current_nxt
            self.age_nxt = 0
            self.power_nxt = 0
            self.voltage_nxt = 0
            self.current_nxt = 0

        return self.voltage_now, self.current_now



from pathlib import Path

import numpy as np
import pytest
from shepherd_core.data_models import CalibrationEmulator as CalEmu
from shepherd_core.data_models import CalibrationHarvester as CalHrv
from shepherd_core.data_models import CalibrationSeries as CalSeries
from shepherd_core.data_models import EnergyDType
from shepherd_core.data_models.task import Compression

from shepherd_core import CalibrationPair
from shepherd_core import Reader
from shepherd_core import Writer


def generate_shp_file(
    store_path: Path,
    mode: str | None = None,
    datatype: str | None = None,
    window_samples: int | None = None,
    cal_data: CalSeries | CalEmu | CalHrv | None = None,
    config: dict | None = None,
    compression: Compression | None = Compression.default,
    hostname: str = "unknown",
) -> Path:
    if config is None:
        config = {}
    with Writer(
        store_path,
        mode=mode,
        datatype=datatype,
        window_samples=window_samples,
        cal_data=cal_data,
        compression=compression,
        verbose=True,
    ) as file:
        file.store_hostname(hostname)
        file.store_config(config)
        duration_s = 2
        timestamps = np.arange(0.0, duration_s, file.sample_interval_ns / 1e9)
        voltages = np.linspace(3.60, 1.90, int(file.samplerate_sps * duration_s))
        currents = np.linspace(100e-6, 2000e-6, int(file.samplerate_sps * duration_s))
        file.append_iv_data_si(timestamps, voltages, currents)
    return store_path


@pytest.fixture
def h5_path(tmp_path: Path) -> Path:
    return tmp_path / "hrv_test.h5"


@pytest.fixture
def h5_file(h5_path: Path) -> Path:
    generate_shp_file(h5_path)
    return h5_path


def test_writer_basics(h5_path: Path) -> None:
    generate_shp_file(h5_path)
    with Reader(h5_path, verbose=True) as file:
        assert round(file.runtime_s) == 2
        assert file.get_window_samples() == 0
        assert file.get_mode() == "harvester"
        assert file.get_config() == {}
        assert file.get_datatype() == EnergyDType.ivsample
        assert file.get_hostname() == "unknown"


def test_writer_compression_1(h5_path: Path) -> None:
    generate_shp_file(h5_path, compression=Compression.gzip1)


def test_writer_unique_path(h5_file: Path) -> None:
    with Writer(h5_file) as sfw:
        assert sfw.file_path != h5_path


def test_writer_faulty_mode(h5_path: Path) -> None:
    with pytest.raises(ValueError):  # noqa: PT011
        generate_shp_file(h5_path, mode="excavator", datatype="ivcurve")


def test_writer_faulty_datatype(h5_path: Path) -> None:
    with pytest.raises(ValueError):  # noqa: PT011
        generate_shp_file(h5_path, mode="emulator", datatype="ivcurve")


def test_writer_faulty_path(tmp_path: Path) -> None:
    with pytest.raises(TypeError):
        generate_shp_file(tmp_path)


def test_writer_init_cal_emu(h5_path: Path) -> None:
    vgain = 0.123
    generate_shp_file(h5_path, cal_data=CalEmu(dac_V_A=CalibrationPair(gain=vgain)))
    # dac_V_A translates to .ds_voltage, should stay untouched
    with Reader(h5_path) as sfr:
        assert sfr.ds_voltage.attrs["gain"] == vgain
    with Writer(h5_path, modify_existing=True) as sfw:
        assert sfw.ds_voltage.attrs["gain"] == vgain


def test_writer_init_cal_hrv(h5_path: Path) -> None:
    vgain = 0.123
    generate_shp_file(h5_path, cal_data=CalHrv(adc_V_Sense=CalibrationPair(gain=vgain)))
    # adc_V_Sense translates to .ds_voltage, should stay untouched
    with Reader(h5_path) as sfr:
        assert sfr.ds_voltage.attrs["gain"] == vgain
    with Writer(h5_path, modify_existing=True) as sfw:
        assert sfw.ds_voltage.attrs["gain"] == vgain


def test_writer_init_cal_series(h5_path: Path) -> None:
    vgain = 0.123
    cs = CalSeries.from_cal(CalHrv(adc_V_Sense=CalibrationPair(gain=vgain)))
    generate_shp_file(h5_path, cal_data=cs)
    # adc_V_Sense translates to .ds_voltage, should stay untouched
    with Reader(h5_path) as sfr:
        assert sfr.ds_voltage.attrs["gain"] == vgain
    with Writer(h5_path, modify_existing=True) as sfw:
        assert sfw.ds_voltage.attrs["gain"] == vgain


def test_writer_faulty_window(h5_path: Path) -> None:
    with pytest.raises(ValueError):  # noqa: PT011
        generate_shp_file(h5_path, mode="harvester", datatype="ivcurve")
    with pytest.raises(ValueError):  # noqa: PT011
        generate_shp_file(
            h5_path,
            mode="harvester",
            datatype="ivcurve",
            window_samples=0,
        )


def test_writer_append_raw(h5_path: Path) -> None:
    with Writer(h5_path) as sfw:
        data_nd = np.zeros((sfw.CHUNK_SAMPLES_N,))
        time_float = 5.5
        time_int = 3
        time_nd = 3 + data_nd
        sfw.append_iv_data_raw(time_float, data_nd, data_nd)
        sfw.append_iv_data_raw(time_int, data_nd, data_nd)
        sfw.append_iv_data_raw(time_nd, data_nd, data_nd)
        with pytest.raises(TypeError):
            sfw.append_iv_data_raw(None, data_nd, data_nd)


def test_writer_align(h5_path: Path) -> None:
    with Writer(h5_path) as sfw:
        length = int(5.5 * sfw.CHUNK_SAMPLES_N)
        time_nd = np.arange(0, length * sfw.sample_interval_ns, sfw.sample_interval_ns)
        data_nd = np.zeros((int(length),))
        sfw.append_iv_data_raw(time_nd, data_nd, data_nd)
    with Reader(h5_path) as sfr:
        assert sfr.ds_voltage.size < length


def test_writer_not_align(h5_path: Path) -> None:
    with Writer(h5_path) as sfw:
        length = int(5.5 * sfw.CHUNK_SAMPLES_N)
        sample_interval = 10 * sfw.sample_interval_ns
        time_nd = np.arange(0, length * sample_interval, sample_interval)
        data_nd = np.zeros((int(length),))
        sfw.append_iv_data_raw(time_nd, data_nd, data_nd)
    with Reader(h5_path) as sfr:
        assert sfr.ds_voltage.size == length


def test_writer_setter(h5_path: Path) -> None:
    name = "pingu"
    with Writer(h5_path) as sfw:
        sfw["hostname"] = name
    with Reader(h5_path) as sfr:
        assert sfr.get_hostname() == name


def test_writer_modify_mode(h5_path: Path) -> None:
    generate_shp_file(h5_path, mode="harvester")
    with Reader(h5_path) as sfr:
        assert sfr.get_mode() == "harvester"
    with Writer(h5_path, modify_existing=True, mode="emulator") as sfw:
        assert sfw.get_mode() == "emulator"
    with Reader(h5_path) as sfr:
        assert sfr.get_mode() == "emulator"


def test_writer_not_modify_mode(h5_path: Path) -> None:
    generate_shp_file(h5_path, mode="harvester")
    with Reader(h5_path) as sfr:
        assert sfr.get_mode() == "harvester"
    with Writer(h5_path, modify_existing=True) as sfw:
        assert sfw.get_mode() == "harvester"
    with Reader(h5_path) as sfr:
        assert sfr.get_mode() == "harvester"


def test_writer_modify_datatype(h5_path: Path) -> None:
    generate_shp_file(h5_path, datatype="ivcurve", window_samples=1024)
    with Reader(h5_path) as sfr:
        assert sfr.get_datatype() == EnergyDType.ivcurve
    with Writer(h5_path, modify_existing=True, datatype="ivsample") as sfw:
        # should not be possible, but who am I?
        assert sfw.get_datatype() == EnergyDType.ivsample
    with Reader(h5_path) as sfr:
        assert sfr.get_datatype() == EnergyDType.ivsample


def test_writer_not_modify_datatype(h5_path: Path) -> None:
    generate_shp_file(h5_path, datatype="ivcurve", window_samples=1024)
    with Reader(h5_path) as sfr:
        assert sfr.get_datatype() == EnergyDType.ivcurve
    with Writer(h5_path, modify_existing=True) as sfw:
        assert sfw.get_datatype() == EnergyDType.ivcurve
    with Reader(h5_path) as sfr:
        assert sfr.get_datatype() == EnergyDType.ivcurve


def test_writer_modify_window_samples(h5_path: Path) -> None:
    generate_shp_file(h5_path, window_samples=1024)
    with Reader(h5_path) as sfr:
        assert sfr.get_window_samples() == 1024
    with Writer(h5_path, modify_existing=True, window_samples=2048) as sfw:
        # should not be possible, but who am I?
        assert sfw.get_window_samples() == 2048
    with Reader(h5_path) as sfr:
        assert sfr.get_window_samples() == 2048


def test_writer_not_modify_window_samples(h5_path: Path) -> None:
    generate_shp_file(h5_path, window_samples=1024)
    with Reader(h5_path) as sfr:
        assert sfr.get_window_samples() == 1024
    with Writer(h5_path, modify_existing=True) as sfw:
        assert sfw.get_window_samples() == 1024
    with Reader(h5_path) as sfr:
        assert sfr.get_window_samples() == 1024


def test_writer_modify_cal_data(h5_path: Path) -> None:
    cal1 = CalSeries.from_cal(CalHrv(adc_V_Sense=CalibrationPair(gain=1.23)))
    cal2 = CalSeries.from_cal(CalHrv(adc_V_Sense=CalibrationPair(gain=2.34)))
    generate_shp_file(h5_path, cal_data=cal1)
    with Reader(h5_path) as sfr:
        assert sfr.get_calibration_data().voltage.gain == 1.23
    with Writer(h5_path, modify_existing=True, cal_data=cal2) as sfw:
        assert sfw.get_calibration_data().voltage.gain == 2.34
    with Reader(h5_path) as sfr:
        assert sfr.get_calibration_data().voltage.gain == 2.34


def test_writer_not_modify_cal_data(h5_path: Path) -> None:
    cal1 = CalSeries.from_cal(CalHrv(adc_V_Sense=CalibrationPair(gain=1.23)))
    generate_shp_file(h5_path, cal_data=cal1)
    with Reader(h5_path) as sfr:
        assert sfr.get_calibration_data().voltage.gain == 1.23
    with Writer(h5_path, modify_existing=True) as sfw:
        assert sfw.get_calibration_data().voltage.gain == 1.23
    with Reader(h5_path) as sfr:
        assert sfr.get_calibration_data().voltage.gain == 1.23


def test_writer_modify_hostname(h5_path: Path) -> None:
    generate_shp_file(h5_path, hostname="bubu")
    with Reader(h5_path) as sfr:
        assert sfr.get_hostname() == "bubu"
    with Writer(h5_path, modify_existing=True) as sfw:
        sfw.store_hostname("sheep")
        assert sfw.get_hostname() == "sheep"
    with Reader(h5_path) as sfr:
        assert sfr.get_hostname() == "sheep"


def test_writer_not_modify_hostname(h5_path: Path) -> None:
    generate_shp_file(h5_path, hostname="bubu")
    with Reader(h5_path) as sfr:
        assert sfr.get_hostname() == "bubu"
    with Writer(h5_path, modify_existing=True) as sfw:
        assert sfw.get_hostname() == "bubu"
    with Reader(h5_path) as sfr:
        assert sfr.get_hostname() == "bubu"


# TODO: get_config

# TODO:
#  - test writing different and confirming them
#  - different compressions and their size relative to each other
#  - also invalid
#  - read raw-data



"""Helper-functions for better support of timezones."""

import time
from datetime import datetime
from datetime import timedelta
from datetime import timezone


def local_tz() -> timezone:
    """Query the local timezone of the user."""
    if time.daylight:
        return timezone(timedelta(seconds=-time.altzone), time.tzname[1])
    return timezone(timedelta(seconds=-time.timezone), time.tzname[0])


def local_now() -> datetime:
    """Query the local date-time of the user."""
    return datetime.now(tz=local_tz())


def local_iso_date() -> str:
    """Query the local date-time of the user as a ISO 8601 date."""
    return local_now().date().isoformat()
